#!/usr/bin/perl

######################################################################
package Cognition::HTMLParser;
######################################################################

use utf8;
use Cognition::Export::Entity;
use Cognition::HTMLParser::DocModel;
use Cognition::HTMLParser::eRDF;
use Cognition::HTMLParser::HTTP;
use Cognition::HTMLParser::RDFa;
use Cognition::HTMLParser::RDF;
use Cognition::HTMLParser::RoleAttr;
use Cognition::HTMLParser::Metadata;
use Cognition::HTMLParser::Utils;
use Cognition::MagicString;
use Cognition::Misc;
use Cognition::Namespace;
use Cognition::RDFModel::Property;
use Cognition::RDFModel::Subject;
use Cognition::RDFModel::Value;
use Cognition::RDFModel::Smusher;
use Cognition::uF::COinS;
use Cognition::uF::RelEnclosure;
use Cognition::uF::RelLicense;
use Cognition::uF::RelTag;
use Cognition::uF::XFN;
use Cognition::uF::adr;
use Cognition::uF::figure;
use Cognition::uF::geo;
use Cognition::uF::hAlarm;
use Cognition::uF::hAtom;
use Cognition::uF::hAudio;
use Cognition::uF::hCalendar;
use Cognition::uF::hCard;
use Cognition::uF::hEvent;
use Cognition::uF::hMeasure;
use Cognition::uF::hRecipe;
use Cognition::uF::hResume;
use Cognition::uF::hReview;
use Cognition::uF::hTodo;
use Cognition::uF::species;
use Cognition::uF::xoxo;
use Encode;
use HTML::Entities qw(:DEFAULT encode_entities_numeric %entity2char);
use HTML::Tagset;
use HTML::TreeBuilder;
use LWP::UserAgent;
use URI::URL;
use XML::LibXML qw(:all);
use strict;

BEGIN
{
	our $VERSION = 0.1;
	our $XDP;
}

sub new
{
	my $class    = shift;
	my $source   = shift;
	my $opts     = shift;
	
	my $this   = {
		'Source'   => $source,
		'UA'       => $opts->{ua},
		'REQUEST'  => $opts->{request},
		'RESPONSE' => $opts->{response},
		'BASE'     => $opts->{base},
		'options'  => {
			'erdf_strict_profiles'  => 1,
			'grddl_fetch'  => 0,
			'grddl_strict_profiles' => 0,
			'normalise_lists'       => 1,
			'nofollow'     => undef,
			'p_AUTO'       => 1,
			'p_comments'   => 1,
			'p_erdf'       => 1,
			'p_grddl'      => 1,
			'p_http'       => 0,
			'p_rdf'        => 1,
			'p_rdfa'       => 1,
			'p_rdfx'       => 1,
			'p_structure'  => 0,
			'p_uf'         => 1,
			'rdfa_strict_doctype'   => 0,
			'rdfa_strict_version'   => 0,
			'rdfa_strings' => 0,
			'smush'        => 0,
			'uf_strict_profiles'    => 0,

		},
	};

	bless $this, $class;
	
	# Process options.
	foreach my $o (keys %$opts)
	{
		$this->{options}->{$o} = $opts->{$o}
			if (defined $opts->{$o} && defined $this->{options}->{$o});
	}
	
	foreach my $h (@{ $this->{options}->{nofollow} })
	{
		$this->{NoFollow}->{$h} = 1;
	}
	
	# Saved configuration stuff.
	$this->load_stored_data('subPropertyOf', 'reasoning');

	# Infrastructure stuff
	$this->build_dom_tree;
	$this->find_base_uri;
	Cognition::HTMLParser::Utils::PrepareXPointers($this->{DOM});
	Cognition::HTMLParser::Utils::PrepareXmlLang($this->{DOM}->documentElement, undef, $opts->{response});
	$this->find_namespaces;
	$this->find_page_profiles;

	# Parse metadata automatically if that option is enabled.
	if ($this->{options}->{p_AUTO})
	{
		$this->parse_metadata;
		$this->parse_uf;
		
		$this->smush;
	}

	return $this;	
}


sub build_dom_tree
{
	my $this = shift;

	# XML::LibXML parser.
	$Cognition::HTMLParser::XDP = XML::LibXML->new();
	
	# Attempt to parse.
	eval {
		$this->{DOM} = $Cognition::HTMLParser::XDP->parse_string($this->{Source});
	};
	
#	# If parsing failed, then 
#	if (! $this->{DOM} )
#	{
#		# Attempt to parse.
#		eval {
#			$this->{DOM} = $Cognition::HTMLParser::XDP->parse_html_string($this->{Source});
#		};
#	}
	
	# If parsing failed, then 
	if (! $this->{DOM} )
	{
		$this->{_raw_source} = $this->{Source};
		
		# Figure out encoding.
		my $enc = 'iso-8859-1';
		if ($this->{RESPONSE}->header('content-type') =~ /\bcharset=[\"\'\s]?([a-z0-9-]+)/i
		|| substr($this->{Source},0,2048) =~ /\bencoding=[\"\'\s]?([a-z0-9-]+)/i 
		|| substr($this->{Source},0,2048) =~ /\bcharset=[\"\'\s]?([a-z0-9-]+)/i )
			{ $enc = lc($1); }
		
		# Decode into UTF-8.
		if ($enc ne 'utf-8')
		{
			$this->{decoded_from} = $enc;
			$this->{Source} = decode($enc, $this->{Source});
		}

		# Try parsing again.
		if (!eval {
			$this->{DOM} = $Cognition::HTMLParser::XDP->parse_string($this->{Source});
		})
		
		# If it still fails, then
		{
			# Last ditch attempt at getting this into usable XHTML...
			$this->{Source} = $this->html2xhtml($this->{Source});
			
			# If we can't parse it yet, then die.
			if (!eval {
				$this->{DOM} = $Cognition::HTMLParser::XDP->parse_string($this->{Source});
			})
			{
				print "COULD NOT PARSE!\n";
				die;
			}
		}
	}	
}


sub find_base_uri
{
	my $this = shift;

	$this->{BASE} = $this->{RESPONSE}->base()
		if ($this->{RESPONSE} && $this->{RESPONSE}->base());

	# HTML <base> element.
	my @bases = $this->{DOM}->getElementsByTagName('base');
	my $base;
	foreach my $b (@bases)
	{
		$base = $b->getAttribute('href')
			if (length $b->getAttribute('href'));
	}
	$this->{BASE} = $this->uri($base)
		if (length $base);
}


sub find_namespaces
# Discover namespaces for metadata conforming to RFC 2731. Note
# that eRDF uses the same format for metadata schemas, so we
# can re-use these for eRDF too.
{
	my $this = shift;
	
	# Pull out XML namespaces defined on the root node. For RDFa, etc we'll
	# also read namespaces from other nodes, but not at this point.
	foreach my $a ($this->{DOM}->getDocumentElement->getAttributes)
	{
		my $k = $a->getName;
		my $v = $a->getValue;
		next unless $k =~ /^xmlns\:/i;
		$k =~ s/^xmlns\://i;
	
		$this->mergeNS($k, $v, "The '$k' Namespace", '', 0);
	}
	
	# Find custom RFC 2731 namespaces. FOAF and other stuff goes here.
	foreach my $l ($this->{DOM}->getElementsByTagName('link'))
	{
		my $rel = $l->getAttribute('rel');
		
		if ($rel =~ /^schema[\.\:]([A-Z_][A-Z0-9_-]*)$/i)
		{
			my $prefix = $1;
			my $url    = $this->uri($l->getAttribute('href'));
			my $title  = $l->getAttribute('title');
			
			$this->mergeNS($prefix, $url, ($title||"The '$prefix' Schema"), undef, 1);
		}
	}

	# Well-known namespaces, some defined so that people can be lazy. Others
	# defined because they're not usually used in an RDF-context, but I thought
	# it might be nice to try.
	$this->mergeNS('_', 'bnode:///', 'Blank Nodes', -1, 0);
	$this->mergeNS('dc', 'http://purl.org/dc/elements/1.1/', 'Dublin Core 1.1', -1, 1);
	$this->mergeNS('dcmitype', 'http://purl.org/dc/dcmitype/', 'Dublin Core Types', -1, 1);
	$this->mergeNS('dcterms', 'http://purl.org/dc/terms/', 'Dublin Core Terms', -1, 1);
	$this->mergeNS('earl', 'http://www.w3.org/WAI/ER/EARL/nmg-strawman#', 'EARL', -1, 0);
	$this->mergeNS('foaf', 'http://xmlns.com/foaf/0.1/', 'Friend of a Friend', -1, 1);
	$this->mergeNS('h5', 'http://buzzword.org.uk/rdf/h5#', 'HTML5 Meta Tags', -1, 0);
	$this->mergeNS('ianalink', 'http://www.iana.org/assignments/relation/', 'IANA Link Types', -1, 0);
	$this->mergeNS('invalid', 'http://invalid.invalid/ns#', 'Undefined Namespace', -1, 0);
	$this->mergeNS('link', 'http://www.w3.org/2006/link#', 'Link Vocab', -1, 0);
	$this->mergeNS('li', 'http://buzzword.org.uk/rdf/list-items/', 'List Items', -1, 0);
	$this->mergeNS('owl', 'http://www.w3.org/2002/07/owl#', 'OWL Web Ontology Language', -1, 0);
	$this->mergeNS('rdf', 'http://www.w3.org/1999/02/22-rdf-syntax-ns#', 'RDF/XML Syntax Specification (Revised)', -1, 0);
	$this->mergeNS('rdfs', 'http://www.w3.org/2000/01/rdf-schema#', 'RDF Vocabulary Description Language 1.0', -1, 0);
	$this->mergeNS('xe', 'http://buzzword.org.uk/rdf/xearl#', 'EARL Extension', -1, 0);
	$this->mergeNS('xhtml', 'http://www.w3.org/1999/xhtml#', 'XHTML', -1, 0);
	$this->mergeNS('xhv', 'http://www.w3.org/1999/xhtml/vocab#', 'XHTML Link Vocabulary', -1, 0);
	$this->mergeNS('xsd', 'http://www.w3.org/2001/XMLSchema#', 'XML Schema', -1, 0);
	
	# NS Squatting!
	$this->mergeNS('iso8601', 'urn:iso:std:iso:8601#', 'ISO 8601', -1, 0);
	$this->mergeNS('openid', 'http://openid.net/specs/openid-authentication-1_1.html#', 'OpenID 1.1', -1, 1);
	$this->mergeNS('service', 'http://www.w3.org/2007/app#', 'Atom Publishing Protocol', -1, 1);
	$this->mergeNS('schema', 'urn:ietf:rfc:2731#', "The 'Schema' Metaschema", -1, 1);
}


sub find_page_profiles
{
	my $this = shift;
	
	foreach my $head ($this->{DOM}->getElementsByTagName('head'))
	{
		my $profile = $head->getAttribute('profile');
		$profile =~ s/(^\s+|\s+$)//g;
		foreach my $url (split /\s+/s, $profile)
		{
			$url = $this->uri($url);
			$this->{Profiles}->{ $url } = 1;
		}
	}
	
	foreach my $link (Cognition::HTMLParser::Utils::searchRel('profile', $this->{DOM}))
	{
		my $url = $this->uri($link->getAttribute('href'));
		$this->{Profiles}->{ $url } = 1;
	}
}


sub get_ua
{
	my $this = shift;
	
	$this->{UA} = LWP::UserAgent->new(
			agent      => $Cognition::Misc::uaString,
			parse_head => 0
		)
		unless( $this->{UA} );
	
	return $this->{UA};
}


sub html2xhtml
{
	my $this = shift;
	my $html = shift;
	
	$html =~ s/\<\!DOCTYPE[^>]*\>//i;
	
	# HTML 5 additional tags, not otherwise supported by HTML::TreeBuilder.
	foreach my $inline (qw(time mark meter progress output ruby rb rt rp rtc rbc))
	{
		$HTML::Tagset::isPhraseMarkup{$inline} = 1;
		$HTML::Tagset::is_Possible_Strict_P_Content{$inline} = 1;
		$HTML::Tagset::isBodyElement{$inline} = 1;
		$HTML::Tagset::isKnown{$inline} = 1;
	}
	$HTML::Tagset::isFormElement{'output'} = 1;
	foreach my $empty (qw(command source event-source)) # embed is already there
	{
		$HTML::Tagset::emptyElement{$empty} = 1;
		$HTML::Tagset::canTighten{$empty} = 1;
		$HTML::Tagset::isPhraseMarkup{$empty} = 1;
		$HTML::Tagset::is_Possible_Strict_P_Content{$empty} = 1;
		$HTML::Tagset::isBodyElement{$empty} = 1;
		$HTML::Tagset::isKnown{$empty} = 1;
	}
	$HTML::Tagset::isHeadElement{'event-source'} = 1;
	$HTML::Tagset::isHeadOrBodyElement{'event-source'} = 1;
	foreach my $object (qw(video audio canvas))
	{
		$HTML::Tagset::is_Possible_Strict_P_Content{$object} = 1;
		$HTML::Tagset::isBodyElement{$object} = 1;
		$HTML::Tagset::isKnown{$object} = 1;
	}
	foreach my $block (qw(section header footer article aside nav figure dialog details datagrid)) # menu is already there
	{
		push @HTML::Tagset::p_closure_barriers, $block;
		$HTML::Tagset::isBodyElement{$block} = 1;
		$HTML::Tagset::isKnown{$block} = 1;
	}
	
	my $t = HTML::TreeBuilder->new();
	$t->implicit_tags(1);
	$t->p_strict(1);
	$t->store_comments(1);
	$t->parse_content($html);
	$t->{'xmlns'} = 'http://www.w3.org/1999/xhtml';
	
	my $dtdPath = __FILE__;
	$dtdPath =~ s|HTMLParser.pm$|ext/sgml-lib/REC-xhtml1-20020801/xhtml1-transitional.dtd|;
	my $xhtml = "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n"
		. "<!DOCTYPE html\n"
		. "\tPUBLIC \"-//W3C//DTD XHTML 1.0 Transitional//EN\"\n"
		. "\t\"$dtdPath\">\n"
		. $t->as_XML(undef, "\t", {
			'area'=>1,
			'base'=>1,
			'basefont'=>1,
			'br'=>1,
			'col'=>1,
			'frame'=>1,
			'hr'=>1,
			'img'=>1,
			'input'=>1,
			'isindex'=>1,
			'link'=>1,
			'meta'=>1,
			'param'=>1,
			'command'=>1,
			'source'=>1,
			'event-source'=>1,
			'embed'=>1
		});

	# XHTML now has some annoying named entities.
	#my %ents = %HTML::Entities::entity2char;
	#foreach my $e (keys %ents)
	#{
	#	my $char = $ents{$e};
	#	$e =~ s/\;$//;
	#	$xhtml =~ s/ \& $e \; / sprintf "&#x%X;", ord($char) /gex;
	#}
	
	return encode_entities($xhtml, "\x80-\xFF");
	
	return $xhtml;
}


sub load_stored_data
{
	my $this = shift;

	OUTER: while (my $fn = shift)
	{
		INNER: foreach my $type (qw(rdf n3 turtle xml))
		{
			my $file = $ENV{'HOME'}.'/.cognition/'.$fn.'.'.$type;
			next INNER unless (-e $file);
			
			open FILE, $file;
			my @lines = <FILE>;
			close FILE;
			my $rdf = join '', @lines;
			Cognition::HTMLParser::RDF::handle(
				$this,
				$rdf,
				'file://'.$file,
				($type eq 'rdf' || $type eq 'xml')?'rdfxml':'turtle',
				'RDF1'
			);
		}
	}

}


sub mergeNS
# Can be passed either an object with a metadata_ns method or
# a list of suggestedPrefix, url, title.
{
	my $this    = shift;
	my $mergeme = shift;
	
	if (ref $mergeme && UNIVERSAL::can($mergeme, 'metadata_ns'))
	{
		my $mns = $mergeme->metadata_ns;
		foreach my $k (keys %$mns)
		{
			$this->mergeNS($k, $mns->{$k}->{nsurl}, $mns->{$k}->{title}, undef, 0);
		}
	}
	
	else
	{
		my $suggestedPrefix = $mergeme;
		my $NS = Cognition::Namespace->new($suggestedPrefix, @_);
		
		my $addIt = 1;
		foreach my $existing (@{$this->{Namespaces}})
		{
			if (Cognition::Namespace::equiv($existing, $NS))
			{
				$addIt = 0;
				last;
			}
		}
		
		push @{$this->{Namespaces}}, $NS
			if ($addIt);

		return $suggestedPrefix;
	}
}


sub mergeSubject
{
	my $this    = shift;
	my $mergeme = shift;
	my $key     = shift || 'RDF';
	my $target;
	
	# Refuse to add boring subjects.
	return unless (defined $mergeme && defined $mergeme->{uris});
	return unless (((keys %{$mergeme->{properties}}) > 1)
			 ||      (1 < scalar(@{ $mergeme->{uris} })));

	if (defined $mergeme && $mergeme->uri =~ m#^http://thing-described-by.org/?\?(.+)$#)
	{
		$mergeme->addVal('http://www.w3.org/2000/01/rdf-schema#isDefinedBy',
			Cognition::RDFModel::Value->new($1, 1));
	}
	
	foreach my $ex (keys %{$this->{$key}})
	{
		my $existing = $this->{$key}->{$ex};
		next unless (defined $existing && (ref $existing eq 'Cognition::RDFModel::Subject'));
		
		if ($existing->equal_to($mergeme))
		{
			$target = $ex;
			last;
		}
	}
	
	if (defined $target)
	{
		$this->{$key}->{$target}->merge($mergeme);
		$this->{$key}->{$target}->{page} = $this;
	}
	else
	{
		$this->{$key}->{ $mergeme->uri } = $mergeme;
		$this->{$key}->{ $mergeme->uri }->{page} = $this;
	}
}


sub parse_metadata
{
	my $this = shift;
	$this->{RDF}->{ $this->uri } = Cognition::RDFModel::Subject->new($this->uri);
	
	no strict 'refs';
	# Note: do HTTP first as we use it to find additional metadata profiles!
	foreach my $module (qw(HTTP Metadata RoleAttr RDFa eRDF DocModel
		RDF::Embedded RDF::Comment RDF::Linked RDF::GRDDL))
	{
		my $is_enabled = 'Cognition::HTMLParser::'.$module.'::is_enabled';
		my $consume    = 'Cognition::HTMLParser::'.$module.'::consume';
		&{$consume}($this) if &{$is_enabled}($this);
	}
	use strict 'refs';
}


sub parse_uf
{
	my $this = shift;
	return unless ($this->{options}->{p_uf});
	
	no strict 'refs';
	foreach my $module (qw(hCard adr geo XFN RelTag RelLicense RelEnclosure
		hCalendar hAtom COinS species hMeasure hReview hResume figure hAudio
		hRecipe))
	{
		my $consume    = 'Cognition::uF::'.$module.'::consume';
		&{$consume}($this);
	}
	use strict 'refs';
	
}


sub smush
{
#	my $this = shift;
#	return unless ($this->{options}->{p_uf});
#	
#	return Cognition::RDFModel::Smusher::smush($this->{RDF}, $this->{RDF1});
}


sub uri
# This function performs two seemingly different duties. Firstly, if passed a
# parameter, treats this as a (possibly relative) URI and returns that same URI
# as an absolute URI, given $this's known base URI. Secondly, if not passed a
# parameter, will return $this's base URI. However, if you reflect on this, you
# will notice that the latter duty derives directly from the former.
{
	my $this  = shift;
	my $param = shift || '';
	my $opts  = shift || {};
	
	if ('XML::LibXML::Element' eq ref $opts)
	{
		my $x = {'element' => $opts};
		$opts = $x;
	}
	
	if ($param =~ /^([a-z][a-z0-9\+\.\-]*)\:/i)
	{
		# seems to be an absolute URI, so can safely return "as is".
		return $param;
	}
	elsif ($opts->{'require-absolute'})
	{
		return undef;
	}
	
	my $url   = url $param, $this->{BASE};
	return $url->abs->as_string;
}


1;

