#!/usr/bin/perl

######################################################################
package Cognition::HTMLParser::RDF;
######################################################################

use Cognition::HTMLParser::Utils;
use Cognition::MagicString;
use HTTP::Request;
use RDF::Redland;
use XML::LibXML qw(:all);
use XML::LibXSLT;

sub bnode_to_uri
{
	my $bnode = shift;
	my $store = shift;
	
	return $bnode->uri->as_string
		if ($bnode->is_resource);

	if (!defined $store->{ $bnode->blank_identifier })
	{
		my $pfx = $bnode->blank_identifier;
		$pfx = s/\d+$//;
		$store->{ $bnode->blank_identifier } = Cognition::HTMLParser::Utils::BNodeURI(undef, $pfx);
	}
	
	return $store->{ $bnode->blank_identifier };
}

sub handle
{
	my $this = shift;
	my $rdf  = shift;
	my $base = shift || $this->{RESPONSE}->base;
	my $type = shift || 'rdfxml';
	my $grph = shift || 'RDF';
	
	my $baseobj = new RDF::Redland::URI("$base");
	my $parser  = new RDF::Redland::Parser($type);
	my $stream;
	
	if (!eval{$stream = $parser->parse_string_as_stream($rdf, $baseobj);})
	{
		# Funny little way of outputting an error message.
		my $S = Cognition::RDFModel::Subject->new($base);
		my $V = Cognition::RDFModel::Value->new("Could not parse as $type.");
		$S->addVal('http://www.w3.org/2000/01/rdf-schema#comment', $V);
		$this->mergeSubject($S);
		return;
	}

	my $bnode_uris = {};

	while ($stream && !$stream->end)	
	{
		my $statement = $stream->current;
		
		my $subj = $statement->subject;
		my $prop = $statement->predicate;
		my $obj  = $statement->object;
		
		next if ($subj->is_literal || $prop->is_literal);
		
		my $subj_uri = bnode_to_uri($subj, $bnode_uris);
		my $prop_uri = bnode_to_uri($prop, $bnode_uris);
		
		my $S = Cognition::RDFModel::Subject->new($subj_uri);
		my $P = $prop_uri;
		my $V;
		
		if ($obj->is_literal)
		{
			$V = Cognition::RDFModel::Value->new($obj->literal_value);
			$V->setDatatype($obj->literal_datatype->as_string)
				if ($obj->literal_datatype);
		}
		else
		{
			$V = Cognition::RDFModel::Value->new(bnode_to_uri($obj, $bnode_uris), 1);
		}

		$S->addVal($P, $V);				
		$this->mergeSubject($S, $grph);	
		
		$stream->next;
	}
	
	if ($type eq 'n3' || $type eq 'turtle')
	{
		while ($rdf =~ / \@prefix \s+ ([A-Za-z0-9_-]*) \s* \: \s* <([^>]+)> \s* \. /gx)
		{
			my $nsurl  = $2;
			my $prefix = length($1) ? $1 : 'p1';
			$this->mergeNS($prefix, $nsurl, "RDF $prefix", -1, 0);
		}
	}
	elsif ($RDF::Redland::VERSION gt '1.00_06_00')
	{
		my %ns = $parser->namespaces_seen;
		foreach my $k (keys %ns)
		{
			$this->mergeNS($k, $ns{$k}->as_string, "RDF $k", -1, 0);
		}
	}
	elsif ($type eq 'rdfxml')
	{
		while ($rdf =~ /xmlns\:?([A-Za-z0-9_-]*)\=\"([^"]*)\"/g)
		{
			my $nsurl  = $2;
			my $prefix = length($1) ? $1 : 'p1';
			$this->mergeNS($prefix, $nsurl, "RDF $prefix", -1, 0);
		}
		while ($rdf =~ /xmlns\:?([A-Za-z0-9_-]*)\=\'([^']*)\'/g)
		{
			my $nsurl  = $2;
			my $prefix = length($1) ? $1 : 'p1';
			$this->mergeNS($prefix, $nsurl, "RDF $prefix", -1, 0);
		}
	}
} #/sub handle_rdf

1;


######################################################################
package Cognition::HTMLParser::RDF::Linked;
######################################################################

use Cognition::Export;
use Cognition::HTMLParser::Utils;
use Clone qw(clone);
use strict;

sub is_enabled
{
	my $this = shift;
	return ($this->{options}->{p_rdfx} > 0);
}

sub consume
{
	my $this = shift;
	my @rels =  allOfObj($this->{RDF}->{ $this->uri }, 'http://www.w3.org/1999/xhtml/vocab#meta');
	push @rels, allOfObj($this->{RDF}->{ $this->uri }, 'http://www.w3.org/2000/01/rdf-schema#seeAlso');
	push @rels, allOfObj($this->{RDF}->{ $this->uri }, 'http://www.iana.org/assignments/relation/meta');

	foreach my $r ( @rels )
	{
		my $uri = URI->new( $r->toString );
		next if ($this->{NoFollow}->{ $uri->host });
	
		my $request  = HTTP::Request->new(GET => "$uri");
		$request->header('Accept', 'application/rdf+xml, '
		                         . 'application/x-turtle;q=0.8, '
		                         . 'application/rdf+turtle;q=0.7, '
		                         . 'application/turtle;q=0.7, '
		                         . 'text/turtle;q=0.7, '
		                         . 'application/xhtml+xml;q=0.6, '
		                         . 'text/html;q=0.5, '
		                         . 'text/n3;q=0.4, '
		                         . 'text/rdf+n3;q=0.3, '
		                         . 'application/xml;q=0.1, '
		                         . 'text/xml;q=0.1'
			);
		
		my $response = $this->get_ua->request($request);

		if ($response->header('content-type') =~ m#^(application/xml|text/xml|application/rdf\+xml)#i)
		{
			Cognition::HTMLParser::RDF::handle($this, $response->content, "$uri", 'rdfxml');
		}

		elsif ($response->header('content-type') =~ m#^(application/(x-)?turtle|text/plain)#i)
		{
			Cognition::HTMLParser::RDF::handle($this, $response->content, "$uri", 'turtle');
		}

		elsif ($response->header('content-type') =~ m#^((text|application)/(rdf\+)?n3)#i)
		{
			# Passing "n3" doesn't seem to work. Try as turtle for now.
			Cognition::HTMLParser::RDF::handle($this, $response->content, "$uri", 'turtle');
		}

		elsif ($response->header('content-type') =~ m#^(text/html|application/xhtml\+html)#i)
		{
			# Set options for a recursive parse.
			my $opts          = clone($this->{options});
			$opts->{ua}       = $this->get_ua;
			$opts->{request}  = $request;
			$opts->{response} = $response;
			$opts->{base}     = "$uri";
			$opts->{p_rdfx}--;
			
			my $external_page = Cognition::HTMLParser->new($response->content, $opts);
			my $turtle        = Cognition::Export::Entity::multiTurtle( $external_page->{RDF} );
			Cognition::HTMLParser::RDF::handle($this, $turtle, "$uri", 'turtle');
		}
	}
} #/sub consume

1;


######################################################################
package Cognition::HTMLParser::RDF::Comment;
######################################################################

sub is_enabled
{
	my $this = shift;
	return $this->{options}->{p_comments};
}

sub consume
{
	my $this = shift;
	my $src = $this->{_raw_source} || $this->{Source};

	while ($src =~ /<!--\s*(.*)\s*-->/sg)
	{
		my $comment = $1;
		Cognition::HTMLParser::RDF::handle($this, $comment)
			if ($comment =~ /^[\s\r\n]*<rdf:RDF/);
	}
}

1;


######################################################################
package Cognition::HTMLParser::RDF::Embedded;
######################################################################

sub is_enabled
{
	my $this = shift;
	return $this->{options}->{p_rdf};
}

sub consume
{
	my $this = shift;
	my @rdfs = $this->{DOM}->getElementsByTagNameNS("http://www.w3.org/1999/02/22-rdf-syntax-ns#", 'RDF');
	foreach my $rdf (@rdfs)
		{ Cognition::HTMLParser::RDF::handle($this, $rdf->toString); }
} #/sub consume

1;


######################################################################
package Cognition::HTMLParser::RDF::GRDDL;
######################################################################

sub is_enabled
{
	return 1;
}

sub consume
{
	my $this = shift;
	my @transformations;

	# Get a list of locally defined transformations.
	if ($this->{Profiles}->{'http://www.w3.org/2003/g/data-view'}
		|| !$this->{options}->{grddl_strict_profiles} )
	{
		my $prop = $this->{RDF}->{ $this->uri }
			->property('http://www.w3.org/2003/g/data-view#transformation');
		if ($prop)
		{
			foreach my $v (@{$prop->valObjects})
			{
				my $uri = URI->new($v->toString);
				next if ($this->{options}->{nofollow}->{ $uri->host });
				
				push @transformations, $v->toString if ($v->type);
			}
		}
	}
	
	# Get a list of transformations defined in the profiles.
	if ($this->{options}->{grddl_fetch})
	{
		foreach my $p (keys %{$this->{Profiles}})
		{
			my $uri = URI->new($p);
			next if ($this->{options}->{nofollow}->{ $uri->host });
		
			my $request  = HTTP::Request->new(GET => $p);
			$request->header('Accept', 'application/xhtml+xml, '
											 . 'text/html;q=0.9');
			my $response = $this->get_ua->request($request);

			if ($response->header('content-type') =~ /(html|xml)/i)
			{
				$this->{Profiles}->{$p} = Cognition::HTMLParser->new(
					$response->content,
					{
						ua        => $this->get_ua,
						request   => $request,
						response  => $response,
						p_AUTO    => 0
					});
				my @trans   = Cognition::HTMLParser::Utils::searchRel(
					'profileTransformation', 
					$this->{Profiles}->{$p}->{DOM}
				);
				foreach my $t (@trans)
				{
					my $url = url $t->getAttribute('href'), $response->base;
					push @transformations, $url->abs->as_string
						unless ($this->{options}->{nofollow}->{ $url->host });
				}
			}
		}
	}
	
	my $attr = Cognition::Namespace::to_curie($this,
		'http://www.w3.org/2003/g/data-view#transformation');
	my $attv = $this->{DOM}->documentElement->getAttribute($attr);
	$attv =~ s/(^\s+|\s+$)//g;
	$attv =~ s/\s+/ /g;
	my @attv = split / /, $attv;
	foreach my $a (@attv)
		{ push @transformations, $this->uri($a); }
	
	foreach my $t (@transformations)
	{
		my $uri = URI->new($t);
		next if ($this->{NoFollow}->{ $uri->host });
	
		my $request  = HTTP::Request->new(GET => $t);
		$request->header('Accept', 'application/xslt+xml, '
										 . 'text/xslt;q=0.9, '
										 . 'text/xsl;q=0.9');
		my $response = $this->get_ua->request($request);

		if ($response->header('content-type') =~ /(xslt|xml)/i)
		{
			my $parser     = XML::LibXML->new();
			my $xslt       = XML::LibXSLT->new();
			my $style_doc  = $parser->parse_string($response->content);
			my $stylesheet = $xslt->parse_stylesheet($style_doc);
			my $results    = $stylesheet->transform($this->{DOM});
			my $rdfxml     = $stylesheet->output_string($results);
			
			Cognition::HTMLParser::RDF::handle($this, $rdfxml);
		}
		
		elsif ($response->header('content-type') =~ /^text\/(css|x\-rdf\+css)$/i)
		{
			# CSS based transform goes here!
		}
	}
}

1;
