#!/usr/bin/perl

######################################################################
package Cognition::Misc;
######################################################################

# Miscellaneous variables and functions mostly used by cognitiond.pl and
# cognition.pl.

use strict;
use utf8;
use Clone qw(clone);
use JSON;

our $defaultBaseURI = 'http://example.com/';
our $version        = '0.1-alpha14';
our $uaString       = "cognition/$version ";
our $releaseDate    = '2008-12-14';

sub url_split
{
	my $url = shift;
	
	if ($url =~ /^([^\#]+)\#subject\((.*)\)$/)
	{
		return ("$1", "$2");
	}
	elsif ($url =~ /^([^\#]+)\#(.*)$/)
	{
		return ("$1", "$1#$2");
	}
	else
	{
		return ($url, undef);
	}
}


sub do_export
{
	my ($output_format, $parser, $subjectUrl) = @_;

	if ($output_format =~ /uf-json/i)
	{
		*UNIVERSAL::TO_JSON = sub {
			my $object = shift;
			
			return "$object"
				if ((ref $object) =~ /^(Cognition..Magic|DateTime)/);

			return undef
				if ((ref $object) =~ /^(XML..LibXML)/);
			
			my $clone  = {};
			foreach my $k (keys %$object)
			{
				$clone->{$k} = $object->{$k}
					unless $k =~ /^(_.+|[A-Z]+|isInCal|isTheRoot|associated_hevent|parent_property_node)$/;
			}
			return $clone;
		};
		my $json = to_json({
			'microformats' => {
				'vcard'           => $parser->{'uF'}->{'hCard'},
				'figure'          => $parser->{'uF'}->{'figure'},
				'adr'             => $parser->{'uF'}->{'adr'},
				'geo'             => $parser->{'uF'}->{'geo'},
				'vcalendar'       => $parser->{'uF'}->{'hCalendar'},
				'rel-license'     => $parser->{'uF'}->{'License'},
				'rel-enclosure'   => $parser->{'uF'}->{'Enclosure'},
				'rel-tag'         => $parser->{'uF'}->{'Tag'},
				'haudio'          => $parser->{'uF'}->{'hAudio'},
				'hmeasure'        => $parser->{'uF'}->{'hMeasure'},
				'hresume'         => $parser->{'uF'}->{'hResume'},
				'hrecipe'         => $parser->{'uF'}->{'hRecipe'},
				'hreview'         => $parser->{'uF'}->{'hReview'},
				'hfeed'           => $parser->{'uF'}->{'hAtom'},
				'xfn'             => $parser->{'uF'}->{'XFN'}
			},
			'parser-info' => {
				'name'            => 'Cognition',
				'version'         => $Cognition::Misc::version,
				'page-url'        => $parser->uri
			}
		}, {
			'utf8'            => 1,
			'pretty'          => 1,
			'allow_nonref'    => 1,
			'allow_blessed'   => 1,
			'convert_blessed' => 1,
		});
		
		return "$json\n";
	}
	
	elsif ($output_format =~ /vcf/i)
	{
		if ($subjectUrl)
			{ return Cognition::Export::Contact::outVCF($parser->{RDF}, $subjectUrl); }
		else
			{ return Cognition::Export::Contact::multiVCF($parser->{RDF}); }
	}
	
	elsif ($output_format =~ /rdf.?json/i)
	{
		if ($subjectUrl)
			{ return Cognition::Export::Entity::outJSON($parser->{RDF}, $subjectUrl); }
		else
			{ return Cognition::Export::Entity::multiJSON($parser->{RDF}); }
	}
	
	elsif ($output_format =~ /turtle/i)
	{
		if ($subjectUrl)
			{ return Cognition::Export::Entity::outTurtle($parser->{RDF}, $subjectUrl); }
		else
			{ return Cognition::Export::Entity::multiTurtle($parser->{RDF}); }
	}
	
	elsif ($output_format =~ /n3/i)
	{
		if ($subjectUrl)
			{ return Cognition::Export::Entity::outN3($parser->{RDF}, $subjectUrl); }
		else
			{ return Cognition::Export::Entity::multiN3($parser->{RDF}); }
	}
	
	elsif ($output_format =~ /trix/i)
	{
		if ($subjectUrl)
			{ return Cognition::Export::Entity::outTriX($parser->{RDF}, $subjectUrl); }
		else
			{ return Cognition::Export::Entity::multiTriX($parser->{RDF}); }
	}
	
	elsif ($output_format =~ /recipebook/i)
	{
		if ($subjectUrl)
			{ return Cognition::Export::Recipe::outRecipeBook($parser->{RDF}, $subjectUrl); }
		else
			{ return Cognition::Export::Recipe::multiRecipeBook($parser->{RDF}); }
	}
	
	elsif ($output_format =~ /kml/i)
	{
		if ($subjectUrl)
			{ return Cognition::Export::Location::outKML($parser->{RDF}, $subjectUrl); }
		else
			{ return Cognition::Export::Location::multiKML($parser->{RDF}); }
	}

	elsif ($output_format =~ /m3u/i)
	{
		if ($subjectUrl)
			{ return Cognition::Export::Recording::outM3U($parser->{RDF}, $subjectUrl); }
		else
			{ return Cognition::Export::Recording::multiM3U($parser->{RDF}); }
	}
		
	elsif ($output_format =~ /detect/i)
	{
		if ($subjectUrl)
			{ return Cognition::Export::Entity::outDetect($parser->{RDF}, $subjectUrl); }
		else
			{ return Cognition::Export::Entity::multiDetect($parser->{RDF}); }
	}
	
	elsif ($output_format =~ /jcard/i)
	{
		if ($subjectUrl)
			{ return Cognition::Export::Contact::outJSON($parser->{RDF}, $subjectUrl); }
		else
			{ return Cognition::Export::Contact::multiJSON($parser->{RDF}); }
	}

	elsif ($output_format =~ /atom/i)
	{
		if ($subjectUrl)
			{ return Cognition::Export::Feed::outAtom($parser->{RDF}, $subjectUrl); }
		else
		{
			foreach my $subj (keys %{$parser->{RDF}})
			{
				if ( Cognition::Export::Feed::is($parser->{RDF}->{$subj}) )
				{
					return Cognition::Export::Feed::outAtom($parser->{RDF}, $subj);
				}
			}
		}
	}
	
	elsif ($output_format =~ /ics/i)
	{
		if ($subjectUrl)
		{
			if (Cognition::Export::Calendar::is($parser->{RDF}, $subjectUrl))
				{ return Cognition::Export::Calendar::outICS($parser->{RDF}, $subjectUrl); }
			elsif (Cognition::Export::CalComponent::is($parser->{RDF}, $subjectUrl))
				{ return Cognition::Export::CalComponent::outICS($parser->{RDF}, $subjectUrl); }
		}
		else
			{ return Cognition::Export::Calendar::multiICS($parser->{RDF}); }
	}
	
	else
	{
		if ($subjectUrl)
			{ return Cognition::Export::Entity::outRDF($parser->{RDF}, $subjectUrl); }
		else
			{ return Cognition::Export::Entity::multiRDF($parser->{RDF}); }
	}
}


1;
